from typing import List, Union
import os
import sys

current_dir = os.path.dirname(os.path.abspath(__file__))

sys.path.insert(1, os.path.join(current_dir, '..', '..', '..'))

from ECHO.model import *

def enum_values(enum_type: EnumType) -> str:
    """EnumType to its value separated by ;"""
    assert enum_type.is_enum_type()
    return ';'.join(enum_type)


def int_values(int_type: IntType) -> str:
    """IntType to min..max"""
    assert int_type.is_int_type()
    return f"{int_type.min}..{int_type.max}"


def struct_values(struct_type: StructType) -> str:
    """StructType to sub_types separated by ,"""
    assert struct_type.is_struct_type()
    sub_domains = []
    for sub_t in struct_type:
        if sub_t.is_enum_type():
            sub_domains.append(enum_values(sub_t))
        elif sub_t.is_int_type():
            sub_domains.append(int_values(sub_t))
        else:
            assert False
    return ','.join(sub_domains)


def title_section(title: str) -> str:
    """decorate section title with %"""
    new_lines = '\n\n'
    comments  = '%'*10
    return new_lines + comments + ' ' + title + ' ' + comments + new_lines


def literal(predicate: Literal) -> str:
    """Literal to asp representation"""
    neg = predicate.negated
    lit_str = ""
    if isinstance(predicate, Literal):
        f_name = f"{predicate.fluent.name}"
        lit_str = f_name
        if len(predicate.args) > 0:
            lit_str += '(' + ','.join(map(param_val_2_asp, predicate.args)) + ')'
        
        if neg:
            return "neg(" + lit_str + ")"

        return lit_str
    assert False

def equality_predicate(predicate: EqualityPredicate) -> str:
    """EqualityPredicate to asp representation"""
    if isinstance(predicate, EqualityPredicate):
        operator = predicate.operator.value
        left_val = param_val_2_asp(predicate.left_operand)
        right_val = param_val_2_asp(predicate.right_operand)
        return left_val + operator + right_val
    assert False
    return ''

def param_val_2_asp(param_val: Union[ArithmeticExpr, int, Variable, str]):

    if isinstance(param_val, str): return param_val
    elif isinstance(param_val, int): return str(param_val)
    elif isinstance(param_val, Variable): return param_val.name.upper()
    elif isinstance(param_val, ArithmeticExpr):
        left_val  = param_val_2_asp(param_val.values[0])
        right_val = param_val_2_asp(param_val.values[1])

        op = param_val.operator.value

        return '(' + left_val + op + right_val + ')'
    else:
        assert False

def to_asp_lines(lines: List[str]) -> str:
    if len(lines) == 0: return ""
    SEP = '.\n'
    return SEP.join(lines)+SEP

def next_alpha(s):
    return chr((ord(s.upper())+1 - 65) % 26 + 65)

def fluent_2_asp(f: Fluent) -> str:

    fluent_name = f.name
    if f.type.is_bool_type():
        return "fluent("+fluent_name+")"
    elif f.type.is_enum_type() or f.type.is_int_type():
        t = f.type
        return f"fluent({fluent_name}(X)):-{t.name}(X)"
    elif f.type.is_struct_type():
        l = []
        var = "A"
        s = f"fluent({fluent_name}("
        for sub_t in f.type:
            l.append((var, sub_t))
            var = next_alpha(var)
        vars = ", ".join([var for var, var_type in l])
        s += vars + ")):-"
        s += ",".join([f"{var_type.name}({var})" for var, var_type in l])
        return s

def action_to_asp(action: IAction, pre: str = None) -> str:
    name = action.name
    variables = action.params
    s = ''
    if len(action.params) == 0:
        s = f'action({name})'
        return s
    else:
        parameters = ','.join(map(param_val_2_asp, action.params))
        s = f'action({name}({parameters}))'
    if len(variables) > 0:
        if pre is None:
            return s + f':-{vars_to_asp(variables)}'
        else:
            return s + f':-{vars_to_asp(variables)}, {pre}'

def action_exec(action: IAction, exec_lit: Predicate) -> str:

    variables = set(action.params + exec_lit.variables)
    body = "" if len(variables) == 0 else ':-' + vars_to_asp(variables)
    if len(action.params) == 0:
        return f'exec({action.name},{literal(exec_lit)})' + body
    else:
        parameters = ','.join(map(param_val_2_asp, action.params))
        return f'exec({action.name}({parameters}),{literal(exec_lit)})' + body

def action_causes(action: IAction, cause_lit: Literal) -> str:

    variables = set(action.params + cause_lit.variables)
    body = "" if len(variables) == 0 else ':-' + vars_to_asp(variables)
    if len(action.params) == 0:
        return f'causes({action.name},{literal(cause_lit)})' + body
    else:
        parameters = ','.join(map(param_val_2_asp, action.params))
        return f'causes({action.name}({parameters}),{literal(cause_lit)})' + body

def vars_to_asp(variables: List[Variable]) -> str:

    print(variables)
    return ",".join([f"{var.type.name}({var.name.upper()})" for var in variables])

def independent_rules() -> List[str]:

    return [
        #OPPOSITE:
        "opposite(F, neg(F)) :- fluent(F)",
        "opposite(neg(F), F) :- fluent(F)",
        "#program step(t)",
        #INERTIA:
        "holds(F,t+1) :- opposite(F,G), holds(F,t), not holds(G, t+1)",
        #"holds(F,t):- not holds(G,t), opposite(F,G)",
        #EXECUTABILITY:
        "not_executable(A,t) :- fluent(F), exec(A,F), not holds(F,t)",
        "not_executable(A,t) :- exec(A,G), opposite(F,G), holds(F,t)",
        "executable(A,t) :- not not_executable(A,t), action(A)",
        "holds(F, t+1) :- executable(A,t), occurs(A,t), causes(A,F)",
        #OCCURS
        "1{occurs(A,t) : action(A)}1",
        ":- action(A), occurs(A,t), not executable(A,t)",
        #LAST STEP
        "#program check(t)",
        ":- goal(F), not holds(F,t+1), query(t)",
        ":- goal(neg(F)), holds(F,t+1), query(t)",
        "#program base"
    ]

def compile_classical_into_asp(problem: HierarchicalGoalNetworkProblem) -> str:


    s =  "%Answer set planning.\n\n"
    s += "%Answer set planning: A Survey. E. Pontelli et al. For a survey.\n"
    s += "%Epistemic Multiagent Reasoning with Collaborative Robots: D. Solda' el al. For a practical use-case.\n\n"

    s += title_section("[[\tPROBLEM DEPENDENT RULES\t]]")
    
    s += title_section('TYPES')
    type_asp_convertion = []
    for t in problem.types:
        if t.is_enum_type() or t.is_int_type() or t.is_struct_type():
            if t.is_enum_type():
                type_asp_convertion.append(f"{t.name}({enum_values(t)})")
            elif t.is_int_type():
                type_asp_convertion.append(f"{t.name}({int_values(t)})")
            #elif t.is_struct_type():
            #    type_asp_convertion.append(f"{t.name}({struct_values(t)})")
                

    s += to_asp_lines(type_asp_convertion)

    s += title_section('FLUENTS')

    fluent_asp = []
    for fluent in problem.fluents:
        fluent_asp.append(fluent_2_asp(fluent))

    s += to_asp_lines(fluent_asp)
    
    s += title_section('INITIALLY')

    init_values = []
    for lit in problem.init_values:
        init_values.append("holds(" + literal(lit) + ", 1)")

    s += to_asp_lines(init_values)

    s += title_section('ACTIONS')

    actions = []
    for action in problem.actions:
        action_equality_conds = []
        for precond in action.precondition:
            if isinstance(precond, EqualityPredicate):
                action_equality_conds.append(equality_predicate(precond))
        if len(action_equality_conds) == 0:
            actions.append(action_to_asp(action))
        else:
            actions.append(action_to_asp(action, ", ".join(action_equality_conds)))

    s += to_asp_lines(actions)

    s += title_section('EXECUTABLE')

    executabilities = []

    for action in problem.actions:
        for precond in action.precondition:
            if isinstance(precond, Literal):
                executabilities.append(action_exec(action, precond))

    s += to_asp_lines(executabilities)

    s += title_section('CAUSES')

    causes = []
    for action in problem.actions:
        for effect in action.effects:
            causes.append(action_causes(action, effect))

    s += to_asp_lines(causes)

    s += title_section('GOALS')
    goals = []
    for goal in problem.goals:
        goals.append('goal(' + literal(goal) + ')')

    s += to_asp_lines(goals)
    
    s += title_section("[[\tPROBLEM INDEPENDENT RULES\t]]")

    s += to_asp_lines(independent_rules())

    return s



#!/usr/bin/env python
# Four spaces as indentation [no tabs]
import os
import json
import subprocess
import re
import csv
from pathlib import Path
import sys
import pprint
import time
import random
import traceback
import logging


from Planners.CaseBasedS1 import caseBased_s1_solver
from Planners.CaseBasedS1 import caseBased_s1_distance
from Planners.PlansformerS1 import plansformer_s1
from Planners.PDDL_parser import classical_parser
from Planners.PDDL_parser import SubgoalCompleteness


# Constants
systemALL = -1;
systemONE = 1
systemTWO = 2

plannerALL = -1;

plannerS1_Dist1 = 1
plannerS1_Dist2 = 2
plannerS1_Random = 3
plannerS1_Combined = 4
plannerS1_Plansformer = 5
plannerS1_JacPlans = 6

plannerS2_1 = 1
plannerS2_2 = 2

output_folder = "Output/"
output_folderPl1 = output_folder + "Pl1/"
scripts_folder = "Scripts/"
dbFolder = "Memory/"
#db_file = "memory.db"
jsonFilename = "cases_classical.json"


# Thresholds
threshold1 = 20 #This threshold represents the minimun number of plan (existence) that needs to be generated from S2 (in the same domain) before accepting S1 solution
threshold2 = 0.8 #The value which current_reward/average_revard must surpass to eploy S1 -- FOR NOW NOT EMPLOYED
threshold3 = 0.9 #This value represents the risk-adversion of the system -- The higher it is the more incline to use S2 the system is
threshold4 = 20 #This is the threshold that is used to set 'M' = 1 when S1 hasn't enough experience,
epsilonS1 = 0.1

maxExperience = 1000

# Global variables seen by all the functions
domainFile = ""
problemFile = ""

domainFileNoPath = ""
domainPath = ""
problemFileNoPath = ""
problemPath = ""

context = ""
solutionS1 = []
confidenceS1 = 0
totalTIME = time.time()
difficulty = 0

# Parsing Variables
domain_name = ""
problem_name = ""
init_States = ""
goal_States = ""
number_of_actions = 0
number_of_predicates = 0

def createFolders():
    Path(dbFolder).mkdir(parents=True,exist_ok=True)
    Path(output_folder).mkdir(parents=True,exist_ok=True)
    Path(output_folderPl1).mkdir(parents=True,exist_ok=True)


def getVarFromFile(filename,varname):
    with open(filename) as myfile:
        for line in myfile:
            if ":"+varname in line:
                line = re.sub(r';.*$', '', line)
                discard, var = line.partition(":"+varname)[::2]
                var = var.replace(")", "")
                myfile.close()
                return var.strip()
        raise Exception('Missing variable named '+ varname)
    raise Exception('Missing file named '+ filename)

def readThreshold(thresholdFile):

    global threshold1
    global threshold2
    global threshold3
    global threshold4
    global epsilonS1

    threshold1 = float(getVarFromFile(thresholdFile,"threshold1"))
    threshold2 = float(getVarFromFile(thresholdFile,"threshold2"))
    threshold3 = float(getVarFromFile(thresholdFile,"threshold3"))
    threshold4 = float(getVarFromFile(thresholdFile,"threshold4"))
    epsilonS1  = float(getVarFromFile(thresholdFile,"epsilonS1"))

#Every Solver stores the solution on a tmp file
def readSolutionFromFile(filename):
    try:
        with open(filename) as myfile:
            for line in myfile:
                if "Solution" in line:
                    if '=' in line:
                        discard, sol = line.partition("=")[::2]
                        res=[]
                        #sol = sol.replace(' ','')
                        sol = sol.replace(';','')
                        sol = sol.replace('\n','')

                        acts = sol.split(',')

                        for act in acts:
                            if act and not act.isspace():
                                res.append(act.strip())

                        myfile.close()
                        return res

        #raise Exception('Missing solution in '+ filename)
        return "noSolution"
    except IOError:
        return "noSolution"

def readTimeFromFile(filename):
    with open(filename) as myfile:
        for line in myfile:
            if "TIMED-OUT" in line:
                return "TO"
            elif "completed the search" or "Search time" in line:
                if ':' in line:
                    discard, ret = line.partition(":")[::2]
                    ret = ret.replace(' ','')
                    ret = ret.replace('\n','')
                    ret = ret.replace('s','')


                    myfile.close()
                    return ret

    #raise Exception('Missing plan in '+ filename)
    return "TO"

def executeS1():

    #New S1###################
    similarity_threshold = 0.2 # If < than this is not returned
    json_path = dbFolder+jsonFilename
    try:
        with open(json_path) as f:
            experience = json.load(f)

        cases = experience["cases"]

        if (plannerS1 == plannerS1_Dist1 or plannerS1 == plannerS1_Dist2):
            #init,goal = getStates.States(problemFile) #reading initial and goal states from problem file

            #returned_list has records of this form <path_to_sol, similarity_score, problem_name>
            sol = ""
            confidence = 0
            returned_list = caseBased_s1_distance.doCaseMatch(cases, (plannerS1-1), domain_name, init_States, goal_States, similarity_threshold)
            if returned_list:
                first_act = True
                for act in returned_list[0][1]:
                    if first_act:
                        sol = act
                        first_act = False
                    else:
                        sol = sol + ", " + act
                confidence = returned_list[0][0]



        elif(plannerS1 == plannerS1_Combined):
            #init_States,goal = getStates.States(problemFile) #reading initial and goal states from problem file

            #returned_list has records of this form <path_to_sol, similarity_score, problem_name>
            sol = ""
            confidence = 0
            returned_list0 = caseBased_s1_distance.doCaseMatch(cases, 0, domain_name, init_States, goal_States, similarity_threshold)
            returned_list1 = caseBased_s1_distance.doCaseMatch(cases, 1, domain_name, init_States, goal_States, similarity_threshold)

            if returned_list0 and returned_list1:
                if (returned_list0[0][0] < returned_list1[0][0]):
                    returned_list = returned_list1
                else:
                    returned_list = returned_list0

                first_act = True
                for act in returned_list[0][1]:
                    if first_act:
                        sol = act
                        first_act = False
                    else:
                        sol = sol + ", " + act
                confidence = returned_list[0][0]

        elif (plannerS1 == plannerS1_Random):
            sol = ""
            confidence, plan = caseBased_s1_solver.randomSolve(cases)
            first_act = True
            for act in plan:
                if first_act:
                    sol = act
                    first_act = False
                else:
                    sol = sol + ", " + act

        elif (plannerS1 == plannerS1_Plansformer):
            try:
                tens_confidence, plan = plansformer_s1.solve(domainFile,problemFile)
            except Exception as e:
                 logging.error(traceback.format_exc())
            str_confidence= re.sub(r'tensor\((.+)\)', r'\1', str(tens_confidence))
            confidence = float(str_confidence)
            sol = ""
            first_act = True
            for act in plan:
                if first_act:
                    sol = act
                    first_act = False
                else:
                    sol = sol + ", " + act


        elif(plannerS1 == plannerS1_JacPlans):
            #init_States,goal = getStates.States(problemFile) #reading initial and goal states from problem file

            #returned_list has records of this form <path_to_sol, similarity_score, problem_name>
            sol = ""
            confidenceCB = 0
            confidence = 0
            returned_list = caseBased_s1_distance.doCaseMatch(cases, 1, domain_name, init_States, goal_States, similarity_threshold)

            if returned_list:
                first_act = True
                solCB = ""
            for act in returned_list[0][1]:
                if first_act:
                    solCB = act
                    first_act = False
                else:
                    solCB = solCB + ", " + act
            confidenceCB = returned_list[0][0]

            tens_confidence, plan = plansformer_s1.solve(domainFile,problemFile)
            str_confidence= re.sub(r'tensor\((.+)\)', r'\1', str(tens_confidence))
            confidencePL = float(str_confidence)
            solPL = ""
            first_act = True
            for act in plan:
                if first_act:
                    solPL = act
                    first_act = False
                else:
                    solPL = solPL + ", " + act

            if(confidenceCB > confidencePL):
                confidence = confidenceCB
                sol = solCB
            else:
                confidence = confidencePL
                sol = solPL



        else:
            raise Exception("The requested System 1 has not been implemented yet.")
    #    solString = s1_planner.s1Solver(domain_name,problem_name,json_path)
    #    solString = solString.replace(";", ",")
        resFile = os.path.splitext(domainFileNoPath)[0]+os.path.splitext(problemFileNoPath)[0]+"S1.tmp"
        out = open(output_folderPl1 + resFile, "w")
        out.write("Solution = " + sol)
        out.close()
        return confidence, resFile;
    except IOError:
        return 0, "";

def solveWithS1():
    confidence, resFile = executeS1()
    solutionS1 = readSolutionFromFile(output_folderPl1 + resFile)
    return solutionS1, confidence

def validateSolution(solution):

    #print("solution is: "+ solution)
    if solution == "noSolution":
        return 0

    stringSolution = ""
    count = 1
    for elem in solution:
        stringSolution += elem
        if count < len(solution):
            stringSolution += ", "
            count +=1


    #print("Execution Line is:  sh ./Planners/EFP/scripts/validate_solution.sh " + instanceNameEFP + " " + stringSolution)
    #Classical
    return SubgoalCompleteness.get_correctness(domainFile,stringSolution,problemFile)


def estimateDifficulty():
    #For now difficulty evaluation that does not consider goal or initial state (Maybe include planning grpah lenght?)

    #domain_name, problem_name, init_States, goal_States, number_of_actions, number_of_predicates = classical_parser.get_details(domainFile,problemFile)
    intersection = [value for value in goal_States if value in init_States]
    diff_fluents = len(goal_States) - len(intersection)

    #print("Difficulty is: " + str(diff_fluents*number_of_actions*pow(2, number_of_predicates)))
    #sys.exit(0)

    return (diff_fluents*number_of_actions*pow(2, number_of_predicates))

def estimateTimeCons(planner,difficulty):
    dbFilename = dbFolder + jsonFilename
    range = 100
    timeSTARTConsumption = 0
    matchCount = 0
    #Controlla IL CSV e leggi valori
    try:
        with open(dbFilename) as db:
            db_reader = csv.reader(db, delimiter=',')
            for row in db_reader:
                if(row[0] == domain_name):
                    if(int(row[1]) == planner):
                        rowDiff = float(row[2])
                        if ((rowDiff < (difficulty + range)) and (rowDiff > (difficulty - range))):
                            matchCount += 1
                            timeSTARTConsumption += float(row[3])

    except IOError:
        matchCount = 0


    if matchCount > 0:
        return (timeSTARTConsumption/matchCount)


    else:
        noMatchConsumption = 5
        return noMatchConsumption


def solveWithS2NoPlan(timeLimit):
    solveWithS2(timeLimit, selectPlannerS2())

def solveWithS2(timeLimit, planner):

    #print("Problem </pro>" + problem_name + "</> solved by System </sys>" + str(systemTWO) + "</> using planner </pla>" + str(planner) +"</>.")
    #sys.exit(0)


    if planner == plannerS2_1:

        head_tail = os.path.split(domainFile)
        domainFileNoPath = head_tail[1]
        domainPath = head_tail[0] + "/"

        head_tail = os.path.split(problemFile)
        problemFileNoPath = head_tail[1]
        problemPath = head_tail[0] + "/"

        result = subprocess.run(['bash','./'+ scripts_folder + 'FASTDOWNWARD_solve.sh', domainFileNoPath, domainPath, problemFileNoPath, problemPath, " " + str(int(timeLimit))+"s"])

        resFilename = os.path.splitext(domainFileNoPath)[0]+os.path.splitext(problemFileNoPath)[0]+".out"
        solutionS2 = readSolutionFromFile("tmp/FastDownward/" + resFilename)
        time = readTimeFromFile("tmp/FastDownward/" + resFilename)

    else:
        raise Exception('Planner '+ str(planner) +' is not a known S2-planner')

    if(time == "TO" or solutionS2 == "noSolution"):
        #return False, timeLimit, None
        print("Problem </pro>" + problem_name + "</> could not be solved by System </sys>" + str(systemTWO) + "</> using planner </pla>" + str(planner) +"</>.")
        return False
    else:
        memorizeSolution(systemTWO, planner, 1.0, float(time), 1.0, solutionS2)
        #return True, float(time), solutionS2

def memorizeSolution(system, planner, confidence, elapsedTime, correctness, solution):


    json_path = dbFolder+jsonFilename

    if not os.path.exists(json_path):
        open(json_path, 'w').close()

    memory_file = open(json_path)

    data = json.load(memory_file)

    if ('cases' not in data.keys()) or (len(data['cases']) == 0):
        index = 0
    else:
        index = max( [int(k) for k in data['cases'].keys()] )

    index += 1

    #jsonformat established here

    data['cases'][str(index)] = {}
    data['cases'][str(index)]['domain_name'] = domain_name
    data['cases'][str(index)]['problem_name'] = problem_name
    data['cases'][str(index)]['difficulty'] = difficulty
    data['cases'][str(index)]['system'] = system
    data['cases'][str(index)]['planner'] = planner
    data['cases'][str(index)]['confidence'] = confidence
    data['cases'][str(index)]['correctness'] = correctness
    data['cases'][str(index)]['solving_time'] = elapsedTime
    data['cases'][str(index)]['total_time'] = time.time()-timeSTART

    #init,goal = getStates.States(problemFile) #reading initial and goal states from problem file

    data['cases'][str(index)]['init'] = init_States
    data['cases'][str(index)]['goal'] = goal_States
    data['cases'][str(index)]['plan'] = solution

    json_object = json.dumps(data,indent=4)

    # writing dictionary data into json file
    with open(json_path,"w") as out:
        out.write(json_object)


    print("The solution of </pro>" + problem_name + "</> is </sol>" + str(solution) + "</> with correctness </cor>" + str(correctness) + "</> and has been found in </tim>" + str(elapsedTime) + "s</> by System </sys>" + str(system) + "</>", end = '')
    if (system != systemONE):
        print(" using planner </pla>" + str(planner) +"</>.")
    else:
        print(" using mode </mod>" + str(plannerS1) +"</>.")
    sys.exit(0)

#----NEW

def countSolvedInstances(system,planner):
    dbFilename = dbFolder + jsonFilename
    matchCount = 0
    totalCorrectness = 0


    # Opening JSON file
    try:
        f = open(dbFilename)
        # returns JSON object as a dictionary
        data = json.load(f)

        index = len(data['cases'])
        count = 0
        while (index > 0) and (count < maxExperience):
            if(data['cases'][str(index)]['domain_name'] == domain_name):
                #planner == systemALL means that we accept all the planners
                if(int(data['cases'][str(index)]['system']) == system or system == systemALL):
                    if(int(data['cases'][str(index)]['planner']) == planner or planner == plannerALL):
                        matchCount += 1
            index -= 1
            count += 1

        return matchCount
    except IOError:
        return 0

def getAvgCorrS1(system,planner,slidingWindow):
    dbFilename = dbFolder + jsonFilename
    matchCount = 0
    totalCorrectness = 0

    # Opening JSON file
    f = open(dbFilename)
    # returns JSON object as a dictionary
    data = json.load(f)

    index = len(data['cases'])
    count = 0
    while (index > 0) and (count < maxExperience):
        if(data['cases'][str(index)]['domain_name'] == domain_name):
            #planner == systemALL means that we accept all the planners
            if(int(data['cases'][str(index)]['system']) == system or system == systemALL):
                if(int(data['cases'][str(index)]['planner']) == planner or planner == plannerALL):
                    if (matchCount < slidingWindow):
                        matchCount += 1
                        totalCorrectness += float(data['cases'][str(index)]['correctness'])
                    else:
                        index = 0
        index -= 1
        count += 1


    if matchCount > 0:
        return (totalCorrectness/matchCount)
    else:
        return 0

def tryS1(plannerS1, solutionS1, confidenceS1, timeS1):
    correctnessS1 = validateSolution(solutionS1)
    if (correctnessS1 >= correctnessCntx):
        memorizeSolution(systemONE, plannerS1, confidenceS1, timeS1, correctnessS1, solutionS1)
    else:
        solveWithS2NoPlan(timeLimitCntx)

def selectPlannerS2():
    planner = plannerS2_1 #By default we use plannerS2_1 -- we use label to indicate the planners

    return planner
#-----------------------------------------------
# Main
#-----------------------------------------------
if __name__ == '__main__':


    timeSTART = time.time()
    createFolders()

    domainFile = sys.argv[1]
    problemFile = sys.argv[2]
    context = sys.argv[3]
    readThreshold(sys.argv[4])
    plannerS1 = int(sys.argv[5])


    #plannerS1_Dist1 = 1
    #plannerS1_Dist2 = 2
    #plannerS1_Random = 3
    #plannerS1_Combined = 4
    #plannerS1_Plansformer = 5
    #plannerS1_JacPlans = 6

    #Some Parsing
    domain_name, problem_name, init_States, goal_States, number_of_actions, number_of_predicates = classical_parser.get_details(domainFile,problemFile)


    #correctnessCntx = float(getVarFromFile(context,"correctness"))
    reduced_risk_adversion = 0.1
    correctnessCntx = threshold3 - reduced_risk_adversion

    #@TODO: REMOVE IS FOR DEBUG
    correctnessCntx = 0.5

    timeLimitCntx = float(getVarFromFile(context,"timelimit"))

    difficulty = estimateDifficulty()

    #Test
    #plannerS2 = selectPlannerS2()
    #solveWithS2(1000000,plannerS2)
    #sys.exit(0)


    ######### S1 metacognitive part

    #Sys2 Baseline
    if (plannerS1 == 0):
        plannerS2 = selectPlannerS2()
        solveWithS2(timeLimitCntx,plannerS2)
        print("Problem </pro>" + problem_name + "</> could not be solved by System </sys>Metacognitive</>.")
        sys.exit(0)


    # AUTOMATICALLY CALL S1
    timeS1 = time.time()
    solutionS1, confidenceS1 = solveWithS1()
    timeS1 = time.time() - timeS1
    M = 1


    # IF NUMBER OF CASES IN JSON FILE (that match the domain) IS > THRESHOLD1 THEN CONTINUE WITH TO S1
    if (countSolvedInstances(systemALL,plannerALL) > threshold1):
        if (countSolvedInstances(systemONE,plannerALL) > threshold4):
            M = 1-getAvgCorrS1(systemONE,plannerALL,threshold4)
        else:
            M = 0

        M = 0

        if (confidenceS1 * (1-M) > threshold3):
            tryS1(plannerS1, solutionS1, confidenceS1, timeS1)


    # ######### S2 metacognitive part
    # # Employ the S2 metacognitive structure
    plannerS2 = selectPlannerS2()
    estimatedTimeS2 = estimateTimeCons(plannerS2, difficulty)
    estimatedCostS2 = sys.maxsize

    remainingTime = timeLimitCntx - (time.time() - timeSTART)
    if(remainingTime - estimatedTimeS2 > 0):
        estimatedCostS2 = estimatedTimeS2 / remainingTime

    if (estimatedCostS2 > 1):
        tryS1(plannerS1, solutionS1, confidenceS1, timeS1)
    else:
        probabilityS1 = (1-threshold3)*epsilonS1
        #random() genrates a number between 0 and 1
        if (probabilityS1 > random.random()):
            tryS1(plannerS1, solutionS1, confidenceS1, timeS1)
        else:
            if (len(solutionS1) > 0):
                correctnessS1 = validateSolution(solutionS1)
                #print("Correctness is: " + str(correctnessS1))
                if (correctnessS1 >= correctnessCntx):
                    if((1-(estimatedCostS2 * (1-threshold3))) > (correctnessS1*(1-M))):
                        if (not solveWithS2(remainingTime,plannerS2)):
                            memorizeSolution(systemONE, plannerS1, confidenceS1, timeS1, correctnessS1, solutionS1)
                    else:
                        memorizeSolution(systemONE, plannerS1, confidenceS1, timeS1, correctnessS1, solutionS1)
                else:
                    solveWithS2(remainingTime,plannerS2)
            else:
                solveWithS2(remainingTime,plannerS2)

    print("Problem </pro>" + problem_name + "</> could not be solved by System </sys>Metacognitive</>.")



from typing import List, Union, TypeVar, Iterator
import re
import os
import sys

from .compiler_to_epddl import compile_into_epddl
from .sofai.Planners.EPDDL.parser import EPDDL_Parser

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(1, os.path.join(CURRENT_DIR, '..', '..', '..'))
from ECHO.model import *


TEMP = 'temp'
DOMAIN_F = os.path.join(CURRENT_DIR, TEMP, 'domain.epddl')
PROBLEM_F = os.path.join(CURRENT_DIR, TEMP, 'problem.epddl')
EFP = os.path.join(CURRENT_DIR, 'sofai',
                   'Planners', 'EFP', 'bin', 'efp.out')
EFP_OUTPUT = os.path.join(CURRENT_DIR, TEMP, 'efp_output.txt')


def solve_mae(mepproblem: MEPlanningProblem) -> Iterator[Instantiated_Action]:

    domain_s, problem_s = compile_into_epddl(mepproblem)
    
    print('>'*30)
    print('>'*30)
    print('>'*30)
    print(domain_s)
    print(problem_s)
    with open(DOMAIN_F, 'w') as domain:
        domain.write(domain_s)
    with open(PROBLEM_F, 'w') as problem:
        problem.write(problem_s)
    parser = EPDDL_Parser()
    parser.parse_domain(DOMAIN_F)  # EX_DOMAIN_F)
    parser.parse_problem(PROBLEM_F)  # EX_PROBLEM_F)
    mAp_f = os.path.join(CURRENT_DIR, TEMP,
                         parser.domain_name + "_" + parser.problem_name + '.tmp')
    file_name = parser.print_EFP(os.path.join(CURRENT_DIR, TEMP), mAp_f)

    print(mAp_f)

    os.system(f'''{EFP} {mAp_f} > {EFP_OUTPUT}''')
    #os.system(f'''{EFP} {mAp_f}  -st KRIPKE_OPT -h SUBGOALS > {EFP_OUTPUT}''')
    #  read mAp plan
    mAp_plan = []
    with open(EFP_OUTPUT) as mAp_solution:
        for line in mAp_solution:
            if 'Solution =' in line:
                print(line)
                s = line[len('Solution =  '):-1]
                mAp_plan = s.split(', ')

    print(mAp_plan)
    if mAp_plan == ['']:
        return zip([],[])

    action_name_plan = []
    #collect action names:
    for action in mAp_plan:
        action_name_plan.append(action.split('_')[0])

    instantiated = []
    #collect instantiations:
    for action in mAp_plan:
        instantiated.append(action.split('_')[1:])

    print(action_name_plan)
    print(instantiated)

    def find_in_domain(action_name: str) -> MEAction:
        dom = mepproblem.actions
        for act in dom:
            if act.name == action_name:
                return act
    maection_plan = [find_in_domain(action_name)
                     for action_name in action_name_plan]

    print(maection_plan)
    return zip(maection_plan, instantiated)



"""Module to solve an EpiCla problem"""
from typing import Tuple, Union, List, Dict
import sys
import os
import string
import random
import copy

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

print(CURRENT_DIR)

#import solvers:
sys.path.insert(1, os.path.join(CURRENT_DIR, '..'))
from mae_epddl_planning import solve_mae
sys.path.insert(1, os.path.join(CURRENT_DIR, '..'))
from answer_set_planning import solve_classical

#import modeling objects:
sys.path.insert(1, os.path.join(CURRENT_DIR, '..', '..', '..'))
from ECHO.model import *


def extract_classical_effects(effects: List[Literal], var_val: Dict[Variable, str]) -> List[Literal]:

    effects = copy.deepcopy(effects)

    for idx, effect in enumerate(effects):
        effects[idx] = effect.instatiate(var_val)

    return effects


def extract_classical_poset(poset: Poset, var_val: Dict[Variable, str]) -> List[Literal]:

    return poset.instatiate(var_val)


def solve_echo(echo: ECHOPlanningProblem) -> List[Tuple[
        Instantiated_Action,
        List[Instantiated_Action]
]]:

    meap_problem = echo.meap_problem
    classical_problem = echo.classical_problem

    echo_plan = []
    epistemic_plan = solve_mae(meap_problem)

    for epistemic_action, instatiated_variables in epistemic_plan:

        if epistemic_action.type is MEActionType.ontic:

            var_val = dict(zip(epistemic_action.params, instatiated_variables))
            if isinstance(classical_problem, HierarchicalGoalNetworkProblem):
                poset = extract_classical_poset(epistemic_action.sub_goals, var_val)
                classical_problem.add_poset(poset)

            else:
                effects = extract_classical_effects(epistemic_action.effects, var_val)
                classical_problem.add_goals(*effects)

            final_state, plan = solve_classical(classical_problem)

            echo_plan.append((epistemic_action, plan))

            if isinstance(classical_problem, HierarchicalGoalNetworkProblem):
                classical_problem.reset_poset()
            else:
                classical_problem.reset_goals()

            classical_problem.reset_initial_values()
            classical_problem.add_initial_values(*final_state)

        else:
            echo_plan.append((epistemic_action, []))

    return echo_plan



"""class for the definition of a planning problem"""
import sys, os
from typing import List, Union, Any, Iterator
from dataclasses import dataclass
import logging
from abc import abstractmethod
from enum import Enum

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(1, current_dir)

from .action import IAction, MEAction
from .fluent import Fluent
from .predicate import Literal, Predicate, Predicate
from .variable import Variable
from .ftype import Type
from .method import Method
from .goal import Goal, Poset

class SolvingClassicalPlanningOption(Enum):  # TODO: to move?
    """Types of classical planning solving problems"""
    PARTIAL_ORDER_PLANNING = 1  # not yet implemented
    ANSWER_SET_PLANNING = 2


@dataclass
class PlanningProblem():
    """definition of a planning problem"""
    _inits: List[Predicate]
    _goals: List[Predicate]
    _types: List[Type]
    _variables: List[Variable]
    _fluents: List[Fluent]

    def __init__(self):
        self._inits: List[Literal] = []
        self._goals: List[Literal] = []
        self._types: List[Type] = []
        self._variables: List[Variable] = []
        self._fluents: List[Fluent] = []
        self._name = 'no_name'

    @property
    def name(self) -> str:
        """domain name"""
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def types(self) -> List[Type]:
        """Returns list of problems types"""
        return self._types

    @property
    def fluents(self) -> List[Fluent]:
        """Returns list of problems fluents"""
        return self._fluents

    @property
    def goals(self) -> List[Predicate]:
        """Returns list of goals of the problem"""
        return self._goals

    @property
    def init_values(self) -> List[Literal]:
        """Returns list of initial values of the problem"""
        return self._inits

    @property
    @abstractmethod
    def actions(self) -> List[Union[IAction, MEAction]]:
        """Returns list of actions involved in the problem"""

    @abstractmethod
    def add_action(self, action: List[Union[IAction, MEAction]]):
        """Add action to the problem"""

    def add_type(self, _type: Type) -> None:
        """Add type to the problem"""
        for _t in self._types:
            if _type == _t:
                logging.info(f"{_type} already inserted")
            elif _t.name == _type.name:
                raise Exception(f"{_t} has name {_t.name} as {_type}")
        self._types.append(_type)

    def add_variable(self, variable: Variable) -> None:
        """Add variable to the problem"""
        if variable.type not in self._types:
            raise Exception(f"{variable.type} not added")
        self._variables.append(variable)

    def add_fluent(self, fluent: Fluent):
        """Add fluent to the problem"""
        if fluent.type not in self._types and not fluent.type.is_bool_type():
            raise Exception(f"{fluent.type} not added")
        self._fluents.append(fluent)

    def add_initial_values(self, *args):
        """Add initial literal to the problem"""
        for arg in args:
            if isinstance(arg, Fluent):
                self._inits.append(Literal(fluent=arg, args=[]))
                continue
            self._inits.append(arg)

    def reset_initial_values(self):
        """Remove initial literals"""
        self._inits = []
            
    def add_goals(self, *args):
        """Add final literal to the problem"""
        for arg in args:
            self._goals.append(arg)

    def reset_goals(self):
        """Remove final literals"""
        self._goals = []

    @staticmethod
    def _stringy(objs: List[Any]) -> List[str]:
        return map(lambda obj: str(obj), objs)

    def __repr__(self):

        sep = "\n\n-----------------------\n\n"
        prob_repr = "Problem domain:"
        prob_repr += sep
        prob_repr += "Types\n"
        prob_repr += "\n".join(self._stringy(self._types))
        prob_repr += sep
        prob_repr += "Variables\n"
        prob_repr += "\n".join(self._stringy(self._variables))
        prob_repr += sep
        prob_repr += "Fluents\n"
        prob_repr += "\n".join(self._stringy(self._fluents))
        prob_repr += sep
        prob_repr += "Actions\n"
        prob_repr += "\n".join(self._stringy(self.actions))
        prob_repr += sep + sep
        prob_repr += "Problem instance:"
        prob_repr += sep
        prob_repr += "Initial values\n"
        prob_repr += "\n".join(self._stringy(self._inits))
        prob_repr += sep
        prob_repr += "Goals\n"
        prob_repr += "\n".join(self._stringy(self._goals))
        return prob_repr


@dataclass
class ClassicalPlanningProblem(PlanningProblem):
    """definition of a classical planning problem"""

    _domain: List[IAction]

    def __init__(self):
        super().__init__()
        self._domain = []

    @property
    def actions(self) -> List[IAction]:
        """Returns list of actions involved in the problem"""
        return self._domain

    def add_action(self, action: IAction):
        """Add action to the problem"""
        #TODO ADD CHECK ON THE FLUENTS: IF THEY ARE ADDED TO THE PROBLEM
        for pre in action.precondition:
            for sub_type in pre.types:
                if sub_type not in self._types and not sub_type.is_bool_type():
                    raise Exception(f"""{sub_type} not added,
                    from {pre} in preconditions""")

        for eff in action.effects:
            for sub_type in eff.types:
                if sub_type not in self._types and not sub_type.is_bool_type():
                    raise Exception(f"""{sub_type} not added,
                    from {eff} in effects""")

        self._domain.append(action)


@dataclass
class HierarchicalGoalNetworkProblem(ClassicalPlanningProblem):
    """definition of a classical planning problem"""

    _domain: List[IAction]
    _methods: List[Method]
    _goals: List[Goal]
    _poset: Poset

    def __init__(self):
        super().__init__()
        self._domain = []
        self._methods = []

    @property
    def methods(self) -> List[Method]:
        """Returns list of actions involved in the problem"""
        return self._methods

    def add_method(self, method: Method):
        """Add action to the problem"""
        #TODO ADD CHECK ON THE FLUENTS: IF THEY ARE ADDED TO THE PROBLEM
        for pre in method.precondition:
            for sub_type in pre.types:
                if sub_type not in self._types and not sub_type.is_bool_type():
                    raise Exception(f"""{sub_type} not added,
                    from {pre} in preconditions""")

        for goal in method.goal_poset.get_goals():
            for literal in goal.literals:
                for sub_type in literal.types:
                    if sub_type not in self._types and not sub_type.is_bool_type():
                        raise Exception(f"""{sub_type} not added,
                        from {eff} in effects""")

        self._methods.append(method)

    def add_poset(self, poset: Poset):
        """Add a poset to satisfy"""
        self._poset = poset

    def add_goal(self, goal: Goal):
        """Add subgoals of the domain"""
        self._goals.append(goal)

    @property
    def initial_poset(self) -> Poset:
        """Add initial poset to the domain"""
        return self._poset

    def reset_poset(self) -> None:
        """Remove initial poset to the domain"""
        self._poset = None

    @property
    def goals(self) -> List[Goal]:
        return self._goals
        
@dataclass(repr=False)
class MEPlanningProblem(PlanningProblem):
    """definition of an epistemic multiagent planning problem"""
    _domain: List[MEAction]

    def __init__(self):
        super().__init__()
        self._domain = []

    @property
    def actions(self) -> List[MEAction]:
        """Returns list of actions involved in the problem"""
        return self._domain

    def add_action(self, action: MEAction):
        """Add action to the problem"""
        for pre in action.precondition:
            for sub_type in pre.types:
                if sub_type not in self._types and not sub_type.is_bool_type():
                    raise Exception(f"""{sub_type} not added,
                    from {pre} in preconditions""")

        for eff in action.effects:
            for sub_type in eff.types:
                if sub_type not in self._types and not sub_type.is_bool_type():
                    raise Exception(f"""{sub_type} not added,
                    from {eff} in effects""")

        self._domain.append(action)

@dataclass(repr=False)
class ECHOPlanningProblem():

    classical_problem: ClassicalPlanningProblem
    meap_problem: MEPlanningProblem



from typing import List, Union, Tuple

from .action import IAction, Instantiated_Action, \
    MEAction, MEActionType
from .fluent import Fluent
from .variable import Variable
from .arithmetic_expression import ArithmeticExpr
from .problem import ClassicalPlanningProblem, MEPlanningProblem, HierarchicalGoalNetworkProblem, ECHOPlanningProblem
from .ftype import Type, BoolType, IntType, EnumType, StructType, AgentType
from .predicate import Predicate, Literal, EqualityPredicate,\
    BeliefPredicate, Forall, eq, neq, B,\
    BooleanPredicate, BooleanOperator, EqualityOperator, When, Forall,\
    conj, disj
from .method import Method
from .goal import Goal, Poset




"""This modul contains Methods classes"""
import sys, os
from dataclasses import dataclass
from typing import List

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(1, current_dir)

from .goal import Goal, Poset
from .variable import Variable
from .predicate import Literal

@dataclass
class Method():
    """A method consist of an ordered set of goals
    This class provide also two methods for
    1) give the set of maximal goals
    2) give the set of goal pair (g1, g2) s.t. g1 < g2"""

    name: str
    params: List[Variable]
    precondition: List[Literal]
    goal_poset: Poset



from typing import Tuple, List, Dict, Union
import clingo
from clingo.application import clingo_main, Application, ApplicationOptions
import copy
import os
import sys

from .classical_asp_encoding import compile_classical_into_asp
from .asp_goal_network_encoding import compile_HGN_into_asp

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(1, os.path.join(current_dir, '..', '..', '..'))

from ECHO.model import *
                                                                                           
MAX_STEP = 30

class Context:
    def id(self, x):
        return x
    def seq(self, x, y):
        return [x, y]

def on_model(m):
    print(m)
    pass


def action_2_instantiated_action(problem: ClassicalPlanningProblem, symbol_action: clingo.Symbol)-> Instantiated_Action:

    action = next(act for act in problem.actions if act.name == symbol_action.name)
    action = copy.deepcopy(action)
    symbol_parameters = symbol_action.arguments
    assert len(action.params) == len(symbol_parameters)
    var_val = {}
    for var, value in zip(action.params, symbol_parameters):
        var_val[var] = int(str(value)) if var.type.is_int_type() else str(value)
    return Instantiated_Action(action, var_val)


def correct_argument(struct_type: Type, arguments: clingo.Symbol) -> Union[str, int]:

    params = []
    for sub_type, argument in zip(struct_type, arguments):
        if sub_type.is_int_type():
            params.append(argument.number)
        elif sub_type.is_enum_type():
            params.append(argument.name)
    return params

def to_problem_literal(fluents: List[Fluent], c_symbol: clingo.Symbol) -> Literal:

    fluent = next(iter([f for f in fluents if f.name == c_symbol.name]))
    if fluent.type.is_bool_type():
        return fluent()
    elif fluent.type.is_int_type():
        return fluent(c_symbol.arguments[0].number)
    elif fluent.type.is_enum_type():
        return fluent(c_symbol.arguments[0].name)
    elif fluent.type.is_struct_type():
        return fluent(*correct_argument(fluent.type, c_symbol.arguments))
    assert False


def solve_classical(problem: Union[ClassicalPlanningProblem, HierarchicalGoalNetworkProblem]) -> Tuple[List[Predicate], List[Instantiated_Action]]:

    if isinstance(problem, HierarchicalGoalNetworkProblem):
        asp_encoding = compile_HGN_into_asp(problem)
    elif isinstance(problem, ClassicalPlanningProblem):
        asp_encoding = compile_classical_into_asp(problem)
        
    else:
        raise Exception(f'{type(problem)} is not handled in asp')

    encoding_file_path = os.path.join(current_dir, 'asp_encoding.lp')
    print(asp_encoding)
    print(encoding_file_path)

    with open(encoding_file_path, 'w') as encoding_file:
        encoding_file.write(asp_encoding)
    
    ctl = clingo.Control()

    ctl.load(encoding_file_path)
    ctl.ground([("base", [])], context=Context())
    ctl.add("check", ["t"], "#external query(t).")

    print("solving phase:")

    model = []
    for step in range(1,MAX_STEP):
        print(f'step number: {step}')
        ctl.ground([("step", [clingo.Number(step)])])
        ctl.ground([("check", [clingo.Number(step)])])
        ctl.assign_external(clingo.Function("query", [clingo.Number(step)]), True)
        with ctl.solve(yield_=True) as handle:
            for m in handle:
                if len(m.symbols(atoms=True)) == 0:
                    print('no module')
                else:
                    model = m.symbols(atoms=True)

        ctl.release_external(clingo.Function("query", [clingo.Number(step)]))
        if len(model) >0:
            break

    final_holds = []
    plan = []

    #get maximum timestamp
    max_timestamp = 1
    for fluent in model:
        if fluent.name == 'holds':
            max_timestamp = max(max_timestamp, fluent.arguments[1].number)
    print(f'maximum timestamp {max_timestamp}')
    for fluent in model:
        if fluent.name == 'holds' and fluent.arguments[1].number == max_timestamp:
            if fluent.arguments[0].name != 'neg' and fluent.positive:

                #here!
                final_holds.append(to_problem_literal(problem.fluents, fluent.arguments[0]))
                #final_holds.append(str(fluent.arguments[0]))
        elif fluent.name == 'occurs' and fluent.positive:
            operation_name = fluent.arguments[0].name
            if operation_name in [act.name for act in problem.actions]:
                action = action_2_instantiated_action(problem, fluent.arguments[0])
                index = int(str(fluent.arguments[1]))
                plan.append((action, index))
    
    plan = [x[0] for x in sorted(plan, key=lambda x : x[1])]
    #print(final_holds, plan)
    return final_holds, plan



import sys, os
from dataclasses import dataclass
from typing import Set, List, Iterator, Tuple, Union, Dict
from copy import deepcopy

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(1, current_dir)

from .predicate import Literal
from .variable import Variable


def convert_elw(el: Union[Variable, str, int],
               var_val: Dict[Variable, Union[str, int]]) -> Union[str, int]:

    if isinstance(el, int):
        return el
    if isinstance(el, str):
        return el
    if isinstance(el, Variable):
        return var_val[el]

@dataclass
class Goal():
    """A goal is a named element, labeled by a set of literals"""
    
    name: str
    arguments: List[Union[Variable, str, int]]
    literals: Set[Literal]

    def __post_init__(self):
        for literal_1 in self.literals:
            for literal_2 in self.literals:
                if literal_1 == (- literal_2):
                    raise Exception(f'Incosistent goal {literal_1} and {literal_2}')

    def instatiate(self, var_val: Dict[Variable, Union[str, int]]) -> 'Goal':

        goal = Goal(name=self.name,
                    arguments=list(map(lambda arg : var_val[arg] if isinstance(arg, Variable)
                                                                 else arg,
                                       self.arguments)),
                    literals=list(map(lambda x : x.instatiate(var_val), self.literals)))
        return goal

    def __hash__(self):
        return hash(self.name)        

    def __call__(self, *args) -> 'Goal':
        var_val = dict(zip(self.arguments, list(*args)))
        return self.instatiate(var_val)

class Poset():
    #Poset for now it is only total order

    representation: List[Goal]

    def __init__(self, representation):
        self.representation = representation

    def add_ordered_elements(ordered_elements: List[Goal]) -> None:
        #only one call handled
        self.representation.append(ordered_elements)

    def get_goals(self) -> Set[Goal]:
        return set(self.representation)
        
    def get_maximal_goals(self) -> Iterator[Goal]:

        yield(self.representation[-1])

    def get_precedence_relations(self) -> Iterator[Tuple[Goal, Goal]]:
        
        for i in range(0, len(self.representation)-1):
            yield(self.representation[i], self.representation[i+1])

    def instatiate(self, var_val: Dict[Variable, Union[str, int]]) -> 'Poset':

        return Poset(list(map(lambda x : x.instatiate(var_val), self.representation)))



"""Set up ECHO library"""

from setuptools import setup, find_packages
from setuptools.command.install import install as DistutilsInstall
from os import system

with open("ECHO/README.org", "r") as ldf:
    long_description = ldf.read()

class EFP_Install(DistutilsInstall):
    def run(self):
        system('cd ECHO/engines/mae_epddl_planning/sofai/Planners/EFP; make; cd -')
        DistutilsInstall.run(self)

setup(
    name='echosys',
    version='0.1.35',
    description='A python library for modeling and solving epistemic multi agents planning problems',
    #long_description = long_description,
    #long_description_content_type = '',
    url='https://github.com/DavideSolda/echosys',
    author='Davide Solda\'',
    author_email='davide.solda@tuwien.ac.at',
    license='GNU General Public License v3.0',
    packages=find_packages(),
    install_requires=['clingo'],
    classifiers=[],
    cmdclass={'install': EFP_Install},
    #include_package_data = True,
    data_files=[('../../ECHO/engines/mae_epddl_planning/sofai/Planners/EFP/bin',
                 ['ECHO/engines/mae_epddl_planning/sofai/Planners/EFP/bin/efp.out'])]
    )






"""
def pretty_print_epica_plan(epicla_plan: List[Union[Instantiated_Action,
                                                    Tuple[Instantiated_Action,
                                                          List[Instantiated_Action]]]]) -> None:
    for action in epicla_plan:
            print(action.name)

"""



"""
module for definition of Variable and ArithmeticExpr
which are the possible arguments of a predicate
"""
import os, sys
from typing import List, Union
from enum import Enum

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(1, current_dir)

from .ftype import IntType, EnumType

class ArithmeticOperator(Enum):
    """Arithmetic operators"""
    Plus = "+"
    Sub = "-"


class ArithmeticExpr():
    """
    class for arithmentic expressions involving variables, for istance
    X + 3 or X + Y
    """
    def __init__(self, op: ArithmeticOperator,
                 values: List[Union['ArithmenticExpr', int, 'Variable']]):

        assert len(values) == 2  # till now only binary arithmentic operators
        self._values = values
        self._op = op
        self._types = []
        #  collect subtypes:
        for value in values:
            if isinstance(value, int):
                continue
            if hasattr(value, 'type'):
                self._types.append(value.type)
            elif hasattr(value, 'types'):
                self._types += value.types
            else:
                raise ValueError(f"{value} as arithmetic term")

        #  collect variables:
        self._variables = []
        for v in values:
            if isinstance(v, type(self)):
                self._variables += v.variables
            elif isinstance(v, int):
                continue
            else:  # Variable
                self._variables.append(v)

    @property
    def values(self) -> List[Union['ArithmenticExpr', int, 'Variable']]:
        return self._values

    @property
    def operator(self):
        return self._op

    @property
    def variables(self):
        return self._variables

    @property
    def types(self):
        return self._types

    def _op_int(self, other, op):
        if is_int_fvalue(other):
            return ArithmeticExpr(op=op, values=[self, other])

        raise Exception("""Sum and sub defined only for
        integer expressions/variables/integers""")

    def __add__(self, other):

        return self._op_int(other, ArithmeticOperator.Plus)

    def __radd__(self, other):

        return self._op_int(other, ArithmeticOperator.Plus)

    def __sub__(self, other):

        return self._op_int(other, ArithmeticOperator.Sub)

    def __rsub__(self, other):

        return self._op_int(other, ArithmeticOperator.Sub)

    def __repr__(self):
        return f"{self.values[0]} {self.operator} {self.values[1]}"


def is_int_fvalue(fvalue) -> bool:
    """Check if fvalue (the fluent parameter) is int"""
    return isinstance(fvalue, (int, ArithmeticExpr)) or \
        type(fvalue).__name__ == 'Variable' and \
        isinstance(fvalue.type, IntType)


def is_enum_fvalue(fvalue) -> bool:
    """Check if fvalue (the fluent parameter) is enum"""
    return isinstance(fvalue, str) or \
        type(fvalue).__name__ == 'Variable' and \
        isinstance(fvalue.type, EnumType)
